import os
import sys
import stat
import json
import shutil
import getpass
import logging
from time import sleep
from pathlib import Path
from threading import Lock
from itertools import chain
from contextlib import suppress
from ansible_runner.interface import run
from subprocess import CalledProcessError

from bbot.core import configurator
from bbot.modules import module_loader
from ..misc import can_sudo_without_password, os_platform

log = logging.getLogger("bbot.core.helpers.depsinstaller")


class DepsInstaller:
    def __init__(self, parent_helper):
        self.parent_helper = parent_helper

        # respect BBOT's http timeout
        http_timeout = self.parent_helper.config.get("http_timeout", 30)
        os.environ["ANSIBLE_TIMEOUT"] = str(http_timeout)

        self.askpass_filename = "sudo_askpass.py"
        self._sudo_password = os.environ.get("BBOT_SUDO_PASS", None)
        if self._sudo_password is None:
            if configurator.bbot_sudo_pass is not None:
                self._sudo_password = configurator.bbot_sudo_pass
            elif can_sudo_without_password():
                self._sudo_password = ""
        self.data_dir = self.parent_helper.cache_dir / "depsinstaller"
        self.parent_helper.mkdir(self.data_dir)
        self.setup_status_cache = self.data_dir / "setup_status.json"
        self.command_status = self.data_dir / "command_status"
        self.parent_helper.mkdir(self.command_status)
        self.setup_status = self.read_setup_status()

        self.no_deps = self.parent_helper.config.get("no_deps", False)
        self.ansible_debug = self.parent_helper.config.get("debug", False)
        self.force_deps = self.parent_helper.config.get("force_deps", False)
        self.retry_deps = self.parent_helper.config.get("retry_deps", False)
        self.ignore_failed_deps = self.parent_helper.config.get("ignore_failed_deps", False)
        self.venv = ""
        if sys.prefix != sys.base_prefix:
            self.venv = sys.prefix

        self.all_modules_preloaded = module_loader.preloaded()

        self.ensure_root_lock = Lock()

    def install(self, *modules):
        self.install_core_deps()
        succeeded = []
        failed = []
        try:
            notified = False
            for m in modules:
                # assume success if we're ignoring dependencies
                if self.no_deps:
                    succeeded.append(m)
                    continue
                # abort if module name is unknown
                if m not in self.all_modules_preloaded:
                    log.verbose(f'Module "{m}" not found')
                    failed.append(m)
                    continue
                preloaded = self.all_modules_preloaded[m]
                # make a hash of the dependencies and check if it's already been handled
                # take into consideration whether the venv or bbot home directory changes
                module_hash = self.parent_helper.sha1(
                    json.dumps(preloaded["deps"], sort_keys=True)
                    + self.venv
                    + str(self.parent_helper.bbot_home)
                    + os.uname()[1]
                ).hexdigest()
                success = self.setup_status.get(module_hash, None)
                dependencies = list(chain(*preloaded["deps"].values()))
                if len(dependencies) <= 0:
                    log.debug(f'No setup to do for module "{m}"')
                    succeeded.append(m)
                    continue
                else:
                    if success is None or (success is False and self.retry_deps) or self.force_deps:
                        if not notified:
                            log.hugeinfo(f"Installing module dependencies. Please be patient, this may take a while.")
                            notified = True
                        log.verbose(f'Installing dependencies for module "{m}"')
                        # get sudo access if we need it
                        if preloaded.get("sudo", False) == True:
                            self.ensure_root(f'Module "{m}" needs root privileges to install its dependencies.')
                        success = self.install_module(m)
                        self.setup_status[module_hash] = success
                        if success or self.ignore_failed_deps:
                            log.debug(f'Setup succeeded for module "{m}"')
                            succeeded.append(m)
                        else:
                            log.warning(f'Setup failed for module "{m}"')
                            failed.append(m)
                    else:
                        if success or self.ignore_failed_deps:
                            log.debug(
                                f'Skipping dependency install for module "{m}" because it\'s already done (--force-deps to re-run)'
                            )
                            succeeded.append(m)
                        else:
                            log.warning(
                                f'Skipping dependency install for module "{m}" because it failed previously (--retry-deps to retry or --ignore-failed-deps to ignore)'
                            )
                            failed.append(m)

        finally:
            self.write_setup_status()

        succeeded.sort()
        failed.sort()
        return succeeded, failed

    def install_module(self, module):
        success = True
        preloaded = self.all_modules_preloaded[module]

        # ansible tasks
        ansible_tasks = preloaded["deps"]["ansible"]
        if ansible_tasks:
            success &= self.tasks(module, ansible_tasks)

        # apt
        deps_apt = preloaded["deps"]["apt"]
        if deps_apt:
            self.apt_install(deps_apt)

        # shell
        deps_shell = preloaded["deps"]["shell"]
        if deps_shell:
            success &= self.shell(module, deps_shell)

        # pip
        deps_pip = preloaded["deps"]["pip"]
        deps_pip_constraints = preloaded["deps"]["pip_constraints"]
        if deps_pip:
            success &= self.pip_install(deps_pip, constraints=deps_pip_constraints)

        return success

    def pip_install(self, packages, constraints=None):
        packages_str = ",".join(packages)
        log.info(f"Installing the following pip packages: {packages_str}")

        command = [sys.executable, "-m", "pip", "install", "--upgrade"] + packages

        if constraints:
            contraints_tempfile = self.parent_helper.tempfile(constraints, pipe=False)
            command.append("--constraint")
            command.append(contraints_tempfile)

        process = None
        try:
            process = self.parent_helper.run(command, check=True)
            message = f'Successfully installed pip packages "{packages_str}"'
            output = process.stdout.splitlines()[-1]
            if output:
                message = output
            log.info(message)
            return True
        except CalledProcessError as err:
            log.warning(f"Failed to install pip packages {packages_str} (return code {err.returncode}): {err.stderr}")
        return False

    def apt_install(self, packages):
        """
        Install packages with the OS's default package manager (apt, pacman, dnf, etc.)
        """
        packages_str = ",".join(packages)
        log.info(f"Installing the following OS packages: {packages_str}")
        args = {"name": packages_str, "state": "present"}  # , "update_cache": True, "cache_valid_time": 86400}
        kwargs = {}
        # don't sudo brew
        if os_platform() != "darwin":
            kwargs = {
                "ansible_args": {
                    "ansible_become": True,
                    "ansible_become_method": "sudo",
                }
            }
        success, err = self.ansible_run(module="package", args=args, **kwargs)
        if success:
            log.info(f'Successfully installed OS packages "{packages_str}"')
        else:
            log.warning(
                f"Failed to install OS packages ({err}). Recommend installing the following packages manually:"
            )
            for p in packages:
                log.warning(f" - {p}")
        return success

    def shell(self, module, commands):
        tasks = []
        for i, command in enumerate(commands):
            command_hash = self.parent_helper.sha1(f"{module}_{i}_{command}").hexdigest()
            command_status_file = self.command_status / command_hash
            if type(command) == str:
                command = {"cmd": command}
            command["cmd"] += f" && touch {command_status_file}"
            tasks.append(
                {
                    "name": f"{module}.deps_shell step {i+1}",
                    "ansible.builtin.shell": command,
                    "args": {"executable": "/bin/bash", "creates": str(command_status_file)},
                }
            )
        success, err = self.ansible_run(tasks=tasks)
        if success:
            log.info(f"Successfully ran {len(commands):,} shell commands")
        else:
            log.warning(f"Failed to run shell dependencies")
        return success

    def tasks(self, module, tasks):
        log.info(f"Running {len(tasks):,} Ansible tasks for {module}")
        success, err = self.ansible_run(tasks=tasks)
        if success:
            log.info(f"Successfully ran {len(tasks):,} Ansible tasks for {module}")
        else:
            log.warning(f"Failed to run Ansible tasks for {module}")
        return success

    def ansible_run(self, tasks=None, module=None, args=None, ansible_args=None):
        _ansible_args = {"ansible_connection": "local"}
        if ansible_args is not None:
            _ansible_args.update(ansible_args)
        module_args = None
        if args:
            module_args = " ".join([f'{k}="{v}"' for k, v in args.items()])
        log.debug(f"ansible_run(module={module}, args={args}, ansible_args={ansible_args})")
        playbook = None
        if tasks:
            for task in tasks:
                if "package" in task:
                    # special case for macos
                    if os_platform() == "darwin":
                        # don't sudo brew
                        task["become"] = False
                        # brew doesn't support update_cache
                        task["package"].pop("update_cache", "")
            playbook = {"hosts": "all", "tasks": tasks}
            log.debug(json.dumps(playbook, indent=2))
        if self._sudo_password is not None:
            _ansible_args["ansible_become_password"] = self._sudo_password
        playbook_hash = self.parent_helper.sha1(str(playbook)).hexdigest()
        data_dir = self.data_dir / (module if module else f"playbook_{playbook_hash}")
        shutil.rmtree(data_dir, ignore_errors=True)
        self.parent_helper.mkdir(data_dir)

        res = run(
            playbook=playbook,
            private_data_dir=str(data_dir),
            host_pattern="localhost",
            inventory={
                "all": {"hosts": {"localhost": _ansible_args}},
            },
            module=module,
            module_args=module_args,
            quiet=not self.ansible_debug,
            verbosity=(3 if self.ansible_debug else 0),
            cancel_callback=lambda: None,
        )

        log.debug(f"Ansible status: {res.status}")
        log.debug(f"Ansible return code: {res.rc}")
        success = res.status == "successful"
        err = ""
        for e in res.events:
            # if self.ansible_debug and not success:
            #    log.debug(json.dumps(e, indent=4))
            if e["event"] == "runner_on_failed":
                err = e["event_data"]["res"]["msg"]
                break
        return success, err

    def read_setup_status(self):
        setup_status = dict()
        if self.setup_status_cache.is_file():
            with open(self.setup_status_cache) as f:
                with suppress(Exception):
                    setup_status = json.load(f)
        return setup_status

    def write_setup_status(self):
        with open(self.setup_status_cache, "w") as f:
            json.dump(self.setup_status, f)

    def ensure_root(self, message=""):
        with self.ensure_root_lock:
            if os.geteuid() != 0 and self._sudo_password is None:
                if message:
                    log.warning(message)
                while not self._sudo_password:
                    # sleep for a split second to flush previous log messages
                    sleep(0.1)
                    password = getpass.getpass(prompt="[USER] Please enter sudo password: ")
                    if self.parent_helper.verify_sudo_password(password):
                        log.success("Authentication successful")
                        self._sudo_password = password
                        configurator.bbot_sudo_pass = password
                    else:
                        log.warning("Incorrect password")

    def install_core_deps(self):
        to_install = set()
        # install custom askpass script
        askpass_src = Path(__file__).resolve().parent / self.askpass_filename
        askpass_dst = self.parent_helper.tools_dir / self.askpass_filename
        shutil.copy(askpass_src, askpass_dst)
        askpass_dst.chmod(askpass_dst.stat().st_mode | stat.S_IEXEC)
        # ensure tldextract data is cached
        self.parent_helper.tldextract("evilcorp.co.uk")
        # command: package_name
        core_deps = {"unzip": "unzip", "curl": "curl"}
        for command, package_name in core_deps.items():
            if not self.parent_helper.which(command):
                to_install.add(package_name)
        if to_install:
            self.ensure_root()
            self.apt_install(list(to_install))



import re
import csv
import string
import logging
import wordninja
from pathlib import Path
from contextlib import suppress
from collections import OrderedDict

from .misc import tldextract, extract_words

log = logging.getLogger("bbot.core.helpers.wordcloud")


class WordCloud(dict):
    def __init__(self, parent_helper, *args, **kwargs):
        self.parent_helper = parent_helper
        self.max_backups = 20

        devops_filename = self.parent_helper.wordlist_dir / "devops_mutations.txt"
        self.devops_mutations = set(self.parent_helper.read_file(devops_filename))

        self.dns_mutator = DNSMutator()

        super().__init__(*args, **kwargs)

    def mutations(
        self, words, devops=True, cloud=True, letters=True, numbers=5, number_padding=2, substitute_numbers=True
    ):
        if isinstance(words, str):
            words = (words,)
        results = set()
        for word in words:
            h = hash(word)
            if not h in results:
                results.add(h)
                yield (word,)
        if numbers > 0:
            if substitute_numbers:
                for word in words:
                    for number_mutation in self.get_number_mutations(word, n=numbers, padding=number_padding):
                        h = hash(number_mutation)
                        if not h in results:
                            results.add(h)
                            yield (number_mutation,)
        for word in words:
            for modifier in self.modifiers(
                devops=devops, cloud=cloud, letters=letters, numbers=numbers, number_padding=number_padding
            ):
                a = (word, modifier)
                b = (modifier, word)
                for _ in (a, b):
                    h = hash(_)
                    if h not in results:
                        results.add(h)
                        yield _

    def modifiers(self, devops=True, cloud=True, letters=True, numbers=5, number_padding=2):
        modifiers = set()
        if devops:
            modifiers.update(self.devops_mutations)
        if cloud:
            modifiers.update(set(self))
        if letters:
            modifiers.update(set(string.ascii_lowercase))
        if numbers > 0:
            modifiers.update(self.parent_helper.gen_numbers(numbers, number_padding))
        return modifiers

    def absorb_event(self, event):
        for word in event.words:
            self.add_word(word)
        if event.scope_distance == 0 and event.type.startswith("DNS_NAME"):
            subdomain = tldextract(event.data).subdomain
            if subdomain and not self.parent_helper.is_ptr(subdomain):
                for s in subdomain.split("."):
                    self.dns_mutator.add_word(s)

    def absorb_word(self, word, ninja=True):
        """
        Use word ninja to smartly split the word,
        e.g. "blacklantern" --> "black", "lantern"
        """
        for w in self.parent_helper.extract_words(word):
            self.add_word(w)

    def add_word(self, word, lowercase=True):
        if lowercase:
            word = word.lower()
        try:
            self[word] += 1
        except KeyError:
            self[word] = 1

    def get_number_mutations(self, base, n=5, padding=2):
        results = set()

        # detects numbers and increments/decrements them
        # e.g. for "base2_p013", we would try:
        # - "base0_p013" through "base12_p013"
        # - "base2_p003" through "base2_p023"
        # limited to three iterations for sanity's sake
        for match in list(self.parent_helper.regexes.num_regex.finditer(base))[-3:]:
            span = match.span()
            before = base[: span[0]]
            after = base[span[-1] :]
            number = base[span[0] : span[-1]]
            numlen = len(number)
            maxnum = min(int("9" * numlen), int(number) + n)
            minnum = max(0, int(number) - n)
            for i in range(minnum, maxnum + 1):
                filled_num = str(i).zfill(numlen)
                results.add(f"{before}{filled_num}{after}")
                if not number.startswith("0"):
                    results.add(f"{before}{i}{after}")

        # appends numbers after each word
        # e.g., for "base_www", we would try:
        # - "base1_www", "base2_www", etc.
        # - "base_www1", "base_www2", etc.
        # limited to three iterations for sanity's sake
        number_suffixes = self.parent_helper.gen_numbers(n, padding)
        for match in list(self.parent_helper.regexes.word_regex.finditer(base))[-3:]:
            span = match.span()
            for suffix in number_suffixes:
                before = base[: span[-1]]
                after = base[span[-1] :]
                # skip if there's already a number
                if len(after) > 1 and not after[0].isdigit():
                    results.add(f"{before}{suffix}{after}")
        # basic cases so we don't miss anything
        for s in number_suffixes:
            results.add(f"{base}{s}")
            results.add(base)

        return results

    def truncate(self, limit):
        new_self = dict(self.json(limit=limit))
        self.clear()
        self.update(new_self)

    def json(self, limit=None):
        cloud_sorted = sorted(self.items(), key=lambda x: x[-1], reverse=True)
        if limit is not None:
            cloud_sorted = cloud_sorted[:limit]
        return OrderedDict(cloud_sorted)

    @property
    def default_filename(self):
        return self.parent_helper.scan.home / f"wordcloud.tsv"

    def save(self, filename=None, limit=None):
        if filename is None:
            filename = self.default_filename
        else:
            filename = Path(filename).resolve()
        try:
            if not self.parent_helper.mkdir(filename.parent):
                log.error(f"Failure creating or error writing to {filename.parent} when saving word cloud")
                return
            if len(self) > 0:
                log.debug(f"Saving word cloud to {filename}")
                with open(str(filename), mode="w", newline="") as f:
                    c = csv.writer(f, delimiter="\t")
                    for word, count in self.json(limit).items():
                        c.writerow([count, word])
                log.debug(f"Saved word cloud ({len(self):,} words) to {filename}")
                return True, filename
            else:
                log.debug(f"No words to save")
        except Exception as e:
            import traceback

            log.warning(f"Failed to save word cloud to {filename}: {e}")
            log.trace(traceback.format_exc())
        return False, filename

    def load(self, filename=None):
        if filename is None:
            wordcloud_path = self.default_filename
        else:
            wordcloud_path = Path(filename).resolve()
        log.verbose(f"Loading word cloud from {wordcloud_path}")
        try:
            with open(str(wordcloud_path), newline="") as f:
                c = csv.reader(f, delimiter="\t")
                for row in c:
                    if len(row) == 1:
                        self.add_word(row[0])
                    elif len(row) == 2:
                        with suppress(Exception):
                            count, word = row
                            count = int(count)
                            self[word] = count
            if len(self) > 0:
                log.success(f"Loaded word cloud ({len(self):,} words) from {wordcloud_path}")
        except Exception as e:
            import traceback

            log_fn = log.debug
            if filename is not None:
                log_fn = log.warning
            log_fn(f"Failed to load word cloud from {wordcloud_path}: {e}")
            if filename is not None:
                log.trace(traceback.format_exc())


class Mutator(dict):
    def mutations(self, words, max_mutations=None):
        mutations = self.top_mutations(max_mutations)
        ret = set()
        if isinstance(words, str):
            words = [words]
        for word in words:
            for m in self.mutate(word, mutations=mutations):
                ret.add("".join(m))
        return ret

    def mutate(self, word, max_mutations=None, mutations=None):
        if mutations is None:
            mutations = self.top_mutations(max_mutations)
        for mutation, count in mutations.items():
            ret = []
            for s in mutation:
                if s is not None:
                    ret.append(s)
                else:
                    ret.append(word)
            yield ret

    def top_mutations(self, n=None):
        if n is not None:
            return dict(sorted(self.items(), key=lambda x: x[-1], reverse=True)[:n])
        else:
            return dict(self)

    def _add_mutation(self, mutation):
        if None not in mutation:
            return
        mutation = tuple([m for m in mutation if m != ""])
        try:
            self[mutation] += 1
        except KeyError:
            self[mutation] = 1

    def add_word(self, word):
        pass


class DNSMutator(Mutator):
    word_regex = re.compile(r"[^_\W]+")
    extract_word_regexes = [
        re.compile(r, re.I)
        for r in [
            r"[a-z]+",
            r"[a-z0-9]+",
            r"[a-z0-9-]+",
            r"[a-z0-9_]+",
            r"[a-z0-9_-]+",
        ]
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        wordlist_dir = Path(__file__).parent.parent.parent / "wordlists"
        wordninja_dns_wordlist = wordlist_dir / "wordninja_dns.txt.gz"
        self.model = wordninja.LanguageModel(wordninja_dns_wordlist)

    def mutations(self, words, max_mutations=None):
        if isinstance(words, str):
            words = [words]
        new_words = set()
        for word in words:
            for e in extract_words(word, acronyms=False, model=self.model, word_regexes=self.extract_word_regexes):
                new_words.add(e)
        return super().mutations(new_words, max_mutations=max_mutations)

    def add_word(self, word):
        for match in self.word_regex.finditer(word):
            start, end = match.span()
            match_str = word[start:end]
            # skip digits
            if match_str.isdigit():
                continue
            before = word[:start]
            after = word[end:]
            basic_mutation = [before, None, after]
            self._add_mutation(basic_mutation)
            match_str_split = self.model.split(match_str)
            if len(match_str_split) > 1:
                for i, s in enumerate(match_str_split):
                    if s.isdigit():
                        continue
                    split_before = "".join(match_str_split[:i])
                    split_after = "".join(match_str_split[i + 1 :])
                    wordninja_mutation = [before + split_before, None, split_after + after]
                    self._add_mutation(wordninja_mutation)



import csv
from contextlib import suppress

from bbot.modules.output.base import BaseOutputModule


class CSV(BaseOutputModule):
    watched_events = ["*"]
    meta = {"description": "Output to CSV"}
    options = {"output_file": ""}
    options_desc = {"output_file": "Output to CSV file"}

    header_row = ["Event type", "Event data", "IP Address", "Source Module", "Scope Distance", "Event Tags"]
    filename = "output.csv"

    def setup(self):
        self.custom_headers = []
        self._headers_set = set()
        self._writer = None
        self._prep_output_dir(self.filename)
        return True

    @property
    def writer(self):
        if self._writer is None:
            self._writer = csv.DictWriter(self.file, fieldnames=self.fieldnames)
            self._writer.writeheader()
        return self._writer

    @property
    def file(self):
        if self._file is None:
            if self.output_file.is_file():
                self.helpers.backup_file(self.output_file)
            self._file = open(self.output_file, mode="a", newline="")
        return self._file

    @property
    def fieldnames(self):
        return self.header_row + list(self.custom_headers)

    def writerow(self, row):
        self.writer.writerow(row)
        self.file.flush()

    def handle_event(self, event):
        # ["Event type", "Event data", "IP Address", "Source Module", "Scope Distance", "Event Tags"]
        self.writerow(
            {
                "Event type": getattr(event, "type", ""),
                "Event data": getattr(event, "data", ""),
                "IP Address": ",".join(
                    str(x) for x in getattr(event, "resolved_hosts", set()) if self.helpers.is_ip(x)
                ),
                "Source Module": str(getattr(event, "module", "")),
                "Scope Distance": str(getattr(event, "scope_distance", "")),
                "Event Tags": ",".join(sorted(list(getattr(event, "tags", [])))),
            }
        )

    def cleanup(self):
        if getattr(self, "_file", None) is not None:
            with suppress(Exception):
                self.file.close()

    def report(self):
        if self._file is not None:
            self.info(f"Saved CSV output to {self.output_file}")

    def add_custom_headers(self, headers):
        if isinstance(headers, str):
            headers = [headers]
        for header in headers:
            if header not in self._headers_set:
                self._headers_set.add(header)
                self.custom_headers.append(header)



from bbot.modules.report.base import BaseReportModule


class aggregate(BaseReportModule):
    flags = ["passive", "safe"]
    meta = {"description": "Summarize statistics at the end of a scan"}

    def report(self):
        for table_row in str(self.scan.stats).splitlines():
            self.info(table_row)



from bbot.modules.shodan_dns import shodan_dns


class zoomeye(shodan_dns):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["affiliates", "subdomain-enum", "passive", "safe"]
    meta = {"description": "Query ZoomEye's API for subdomains", "auth_required": True}
    options = {"api_key": "", "max_pages": 20, "include_related": False}
    options_desc = {
        "api_key": "ZoomEye API key",
        "max_pages": "How many pages of results to fetch",
        "include_related": "Include domains which may be related to the target",
    }

    base_url = "https://api.zoomeye.org"

    def setup(self):
        self.max_pages = self.config.get("max_pages", 20)
        self.headers = {"API-KEY": self.config.get("api_key", "")}
        self.include_related = self.config.get("include_related", False)
        return super().setup()

    def ping(self):
        r = self.helpers.request(f"{self.base_url}/resources-info", headers=self.headers)
        assert int(r.json()["quota_info"]["remain_total_quota"]) > 0, "No quota remaining"

    def handle_event(self, event):
        query = self.make_query(event)
        results = self.query(query)
        if results:
            for hostname in results:
                if hostname == event:
                    continue
                tags = []
                if not hostname.endswith(f".{query}"):
                    tags = ["affiliate"]
                self.emit_event(hostname, "DNS_NAME", event, tags=tags)

    def query(self, query):
        query_type = 0 if self.include_related else 1
        url = f"{self.base_url}/domain/search?q={self.helpers.quote(query)}&type={query_type}&page=" + "{page}"
        for i, j in enumerate(self.helpers.api_page_iter(url, headers=self.headers)):
            results = list(self.parse_results(j))
            if results:
                yield from results
            if not results or i >= (self.max_pages - 1) or self.scan.stopping:
                break

    def parse_results(self, r):
        for entry in r.get("list", []):
            yield entry["name"]



import importlib
from pathlib import Path

from .base import BaseCloudProvider

# dynamically load cloud provider modules
provider_list = []
for file in Path(__file__).parent.glob("*.py"):
    if not file.stem in ("base", "__init__"):
        import_path = f"bbot.core.helpers.cloud.{file.stem}"
        module_variables = importlib.import_module(import_path, "bbot")
        for variable in module_variables.__dict__.keys():
            value = getattr(module_variables, variable)
            if hasattr(value, "__mro__") and not value == BaseCloudProvider and BaseCloudProvider in value.__mro__:
                provider_list.append(value)


class CloudProviders:
    def __init__(self, parent_helper):
        self.parent_helper = parent_helper
        self.providers = {}
        for provider_class in provider_list:
            provider_name = str(provider_class.__name__).lower()
            provider = provider_class(self.parent_helper)
            self.providers[provider_name] = provider
            setattr(self, provider_name, provider)

    def excavate(self, event, http_body):
        for provider in self.providers.values():
            provider.excavate(event, http_body)

    def __iter__(self):
        yield from self.providers.values()



from ..bbot_fixtures import *  # noqa: F401
from bbot.scanner import Scanner


def test_curl(bbot_httpserver, bbot_config):
    scan = Scanner("127.0.0.1", config=bbot_config)
    helpers = scan.helpers
    url = bbot_httpserver.url_for("/curl")
    bbot_httpserver.expect_request(uri="/curl").respond_with_data("curl_yep")
    bbot_httpserver.expect_request(uri="/index.html").respond_with_data("curl_yep_index")
    assert helpers.curl(url=url) == "curl_yep"
    assert helpers.curl(url=url, ignore_bbot_global_settings=True) == "curl_yep"
    assert helpers.curl(url=url, head_mode=True).startswith("HTTP/")
    assert helpers.curl(url=url, raw_body="body") == "curl_yep"
    assert (
        helpers.curl(
            url=url,
            raw_path=True,
            headers={"test": "test", "test2": ["test2"]},
            ignore_bbot_global_settings=False,
            post_data={"test": "test"},
            method="POST",
            cookies={"test": "test"},
            path_override="/index.html",
        )
        == "curl_yep_index"
    )
    # test custom headers
    bbot_httpserver.expect_request("/test-custom-http-headers-curl", headers={"test": "header"}).respond_with_data(
        "curl_yep_headers"
    )
    headers_url = bbot_httpserver.url_for("/test-custom-http-headers-curl")
    curl_result = helpers.curl(url=headers_url)
    assert curl_result == "curl_yep_headers"



from .crobat import crobat


class crt(crobat):
    flags = ["subdomain-enum", "passive", "safe"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {"description": "Query crt.sh (certificate transparency) for subdomains"}

    base_url = "https://crt.sh"
    reject_wildcards = False

    def setup(self):
        self.cert_ids = set()
        return super().setup()

    def request_url(self, query):
        params = {"q": f"%.{query}", "output": "json"}
        url = self.helpers.add_get_params(self.base_url, params).geturl()
        return self.request_with_fail_count(url, timeout=self.http_timeout + 10)

    def parse_results(self, r, query):
        j = r.json()
        for cert_info in j:
            if not type(cert_info) == dict:
                continue
            cert_id = cert_info.get("id")
            if cert_id:
                if hash(cert_id) not in self.cert_ids:
                    self.cert_ids.add(hash(cert_id))
                    domain = cert_info.get("name_value")
                    if domain:
                        for d in domain.splitlines():
                            yield d.lower()



from ..bbot_fixtures import *  # noqa: F401


def test_cloud_helpers(monkeypatch, bbot_scanner, bbot_config):
    scan1 = bbot_scanner("127.0.0.1", config=bbot_config)
    scan1.load_modules()
    aws_event1 = scan1.make_event("amazonaws.com", source=scan1.root_event)
    aws_event2 = scan1.make_event("asdf.amazonaws.com", source=scan1.root_event)
    aws_event3 = scan1.make_event("asdfamazonaws.com", source=scan1.root_event)
    providers = scan1.helpers.cloud.providers

    provider_names = ("aws", "gcp", "azure", "firebase", "digitalocean")

    # make sure they're all here
    for name in provider_names:
        assert name in providers

    # make sure tagging is working
    aws = providers["aws"]
    aws.tag_event(aws_event1)
    assert "cloud-aws" in aws_event1.tags
    aws.tag_event(aws_event2)
    assert "cloud-aws" in aws_event2.tags
    aws.tag_event(aws_event3)
    assert "cloud-aws" not in aws_event3.tags



from pathlib import Path
from omegaconf import OmegaConf

from ..helpers.misc import mkdir
from ..errors import ConfigLoadError
from ..helpers.logger import log_to_stderr

config_dir = (Path.home() / ".config" / "bbot").resolve()
defaults_filename = (Path(__file__).parent.parent.parent / "defaults.yml").resolve()
mkdir(config_dir)
config_filename = (config_dir / "bbot.yml").resolve()
secrets_filename = (config_dir / "secrets.yml").resolve()
default_config = None


def _get_config(filename, name="config", notify=True):
    filename = Path(filename).resolve()
    try:
        conf = OmegaConf.load(str(filename))
        if notify:
            log_to_stderr(f"Loaded {name} from {filename}")
        return conf
    except Exception as e:
        if filename.exists():
            raise ConfigLoadError(f"Error parsing config at {filename}:\n\n{e}")
        return OmegaConf.create()


def get_config():
    global default_config
    default_config = _get_config(defaults_filename, name="defaults")
    return OmegaConf.merge(
        default_config,
        _get_config(config_filename, name="config"),
        _get_config(secrets_filename, name="secrets"),
    )



import os
import logging
from pathlib import Path
from threading import Lock

from . import misc
from .dns import DNSHelper
from .diff import HttpCompare
from .wordcloud import WordCloud
from .cloud import CloudProviders
from .interactsh import Interactsh
from .threadpool import as_completed
from ...scanner.target import Target
from ...modules.base import BaseModule
from .depsinstaller import DepsInstaller


log = logging.getLogger("bbot.core.helpers")


class ConfigAwareHelper:
    from .web import wordlist, request, download, api_page_iter, curl
    from .cache import cache_get, cache_put, cache_filename, is_cached, CacheDict
    from .command import run, run_live, _prepare_command_kwargs, tempfile, feed_pipe, _feed_pipe, tempfile_tail
    from . import ntlm
    from . import regexes
    from . import validators

    def __init__(self, config, scan=None):
        self.config = config
        self._scan = scan
        self.bbot_home = Path(self.config.get("home", "~/.bbot")).expanduser().resolve()
        self.cache_dir = self.bbot_home / "cache"
        self.temp_dir = self.bbot_home / "temp"
        self.tools_dir = self.bbot_home / "tools"
        self.lib_dir = self.bbot_home / "lib"
        self.scans_dir = self.bbot_home / "scans"
        self.wordlist_dir = Path(__file__).parent.parent.parent / "wordlists"
        self.current_dir = Path.cwd()
        self.keep_old_scans = self.config.get("keep_scans", 20)
        self.mkdir(self.cache_dir)
        self.mkdir(self.temp_dir)
        self.mkdir(self.tools_dir)
        self.mkdir(self.lib_dir)
        self._futures = set()
        self._future_lock = Lock()

        self.dns = DNSHelper(self)
        self.depsinstaller = DepsInstaller(self)
        self.word_cloud = WordCloud(self)
        self.dummy_modules = {}

        # cloud helpers
        self.cloud = CloudProviders(self)

    def interactsh(self):
        return Interactsh(self)

    def http_compare(self, url, allow_redirects=False, include_cache_buster=True):
        return HttpCompare(url, self, allow_redirects=allow_redirects, include_cache_buster=include_cache_buster)

    def temp_filename(self):
        """
        temp_filename() --> Path("/home/user/.bbot/temp/pgxml13bov87oqrvjz7a")
        """
        return self.temp_dir / self.rand_string(20)

    def clean_old_scans(self):
        _filter = lambda x: x.is_dir() and self.regexes.scan_name_regex.match(x.name)
        self.clean_old(self.scans_dir, keep=self.keep_old_scans, filter=_filter)

    def make_target(self, events):
        return Target(self.scan, *events)

    @property
    def scan(self):
        if self._scan is None:
            from bbot.scanner import Scanner

            self._scan = Scanner()
        return self._scan

    @property
    def scan_stopping(self):
        return getattr(self.scan, "stopping", False)

    @property
    def in_tests(self):
        return os.environ.get("BBOT_TESTING", "") == "True"

    @staticmethod
    def as_completed(*args, **kwargs):
        return as_completed(*args, **kwargs)

    def _make_dummy_module(self, name, _type="scan"):
        """
        Construct a dummy module, for attachment to events
        """
        try:
            return self.dummy_modules[name]
        except KeyError:
            dummy = DummyModule(scan=self.scan, name=name, _type=_type)
            self.dummy_modules[name] = dummy
            return dummy

    def __getattribute__(self, attr):
        """
        Allow static functions from sub-helpers to be accessed from the main class
        """
        try:
            # first try self
            return super().__getattribute__(attr)
        except AttributeError:
            try:
                # then try misc
                return getattr(misc, attr)
            except AttributeError:
                try:
                    # then try dns
                    return getattr(self.dns, attr)
                except AttributeError:
                    # then die
                    raise AttributeError(f'Helper has no attribute "{attr}"')


class DummyModule(BaseModule):
    _priority = 4

    def __init__(self, *args, **kwargs):
        self._name = kwargs.pop("name")
        self._type = kwargs.pop("_type")
        super().__init__(*args, **kwargs)



from .crobat import crobat


class leakix(crobat):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query leakix.net for subdomains"}

    base_url = "https://leakix.net"

    def request_url(self, query):
        url = f"{self.base_url}/api/subdomains/{self.helpers.quote(query)}"
        return self.request_with_fail_count(url, headers={"Accept": "application/json"})

    def parse_results(self, r, query=None):
        json = r.json()
        if json:
            for entry in json:
                subdomain = entry.get("subdomain", "")
                if subdomain:
                    yield subdomain



import sys

from bbot.modules.base import BaseModule


"""
wrapper for https://github.com/defparam/smuggler.git
"""


class smuggler(BaseModule):
    watched_events = ["URL"]
    produced_events = ["FINDING"]
    flags = ["active", "aggressive", "slow", "web-thorough"]
    meta = {"description": "Check for HTTP smuggling"}

    in_scope_only = True

    deps_ansible = [
        {
            "name": "Get smuggler repo",
            "git": {"repo": "https://github.com/defparam/smuggler.git", "dest": "#{BBOT_TOOLS}/smuggler"},
        }
    ]

    def setup(self):
        self.scanned_hosts = set()
        return True

    def handle_event(self, event):
        host = f"{event.parsed.scheme}://{event.parsed.netloc}/"
        host_hash = hash(host)
        if host_hash in self.scanned_hosts:
            self.debug(f"Host {host} was already scanned, exiting")
            return
        else:
            self.scanned_hosts.add(host_hash)

        command = [
            sys.executable,
            f"{self.scan.helpers.tools_dir}/smuggler/smuggler.py",
            "--no-color",
            "-q",
            "-u",
            event.data,
        ]
        for f in self.helpers.run_live(command):
            if "Issue Found" in f:
                technique = f.split(":")[0].rstrip()
                text = f.split(":")[1].split("-")[0].strip()
                description = f"[HTTP SMUGGLER] [{text}] Technique: {technique}"
                self.emit_event(
                    {"host": str(event.host), "url": event.data, "description": description}, "FINDING", source=event
                )



from bbot.modules.base import BaseModule


class crobat(BaseModule):
    """
    A typical free API-based subdomain enumeration module
    Inherited by several other modules including sublist3r, dnsdumpster, etc.
    """

    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    # tag "subdomain-enum" removed 2023-02-24 because API is offline
    flags = ["passive", "safe"]
    meta = {"description": "Query Project Crobat for subdomains"}

    base_url = "https://sonar.omnisint.io"
    # set module error state after this many failed requests in a row
    abort_after_failures = 5
    # whether to reject wildcard DNS_NAMEs
    reject_wildcards = "strict"
    # this helps combat rate limiting by ensuring that a query doesn't execute
    # until the queue is ready to receive its results
    _qsize = 1

    def setup(self):
        self.processed = set()
        self.http_timeout = self.scan.config.get("http_timeout", 10)
        self._failures = 0
        return True

    def _is_wildcard(self, query):
        for domain, wildcard_rdtypes in self.helpers.is_wildcard_domain(query).items():
            if any(t in wildcard_rdtypes for t in ("A", "AAAA", "CNAME")):
                return True
        return False

    def filter_event(self, event):
        """
        This filter_event is used across many modules
        """
        query = self.make_query(event)
        # reject if already processed
        if self.already_processed(query):
            return False, "Event was already processed"
        eligible, reason = self.eligible_for_enumeration(event)
        if eligible:
            self.processed.add(hash(query))
            return True, reason
        return False, reason

    def eligible_for_enumeration(self, event):
        query = self.make_query(event)
        # check if wildcard
        is_wildcard = self._is_wildcard(query)
        # check if cloud
        is_cloud = False
        if any(t.startswith("cloud-") for t in event.tags):
            is_cloud = True
        # reject if it's a cloud resource and not in our target
        if is_cloud and event not in self.scan.target:
            return False, "Event is a cloud resource and not a direct target"
        # optionally reject events with wildcards / errors
        if self.reject_wildcards:
            if any(t in event.tags for t in ("a-error", "aaaa-error")):
                return False, "Event has a DNS resolution error"
            if self.reject_wildcards == "strict":
                if is_wildcard:
                    return False, "Event is a wildcard domain"
            elif self.reject_wildcards == "cloud_only":
                if is_wildcard and is_cloud:
                    return False, "Event is both a cloud resource and a wildcard domain"
        return True, ""

    def already_processed(self, hostname):
        for parent in self.helpers.domain_parents(hostname, include_self=True):
            if hash(parent) in self.processed:
                return True
        return False

    def abort_if(self, event):
        # this helps weed out unwanted results when scanning IP_RANGES and wildcard domains
        if "in-scope" not in event.tags:
            return True
        if self._is_wildcard(event.data):
            return True
        return False

    def handle_event(self, event):
        query = self.make_query(event)
        results = self.query(query)
        if results:
            for hostname in set(results):
                if hostname:
                    try:
                        hostname = self.helpers.validators.validate_host(hostname)
                    except ValueError as e:
                        self.verbose(e)
                        continue
                    if hostname and hostname.endswith(f".{query}") and not hostname == event.data:
                        self.emit_event(hostname, "DNS_NAME", event, abort_if=self.abort_if)

    def request_url(self, query):
        url = f"{self.base_url}/subdomains/{self.helpers.quote(query)}"
        return self.request_with_fail_count(url)

    def make_query(self, event):
        if "target" in event.tags:
            query = str(event.data)
        else:
            query = self.helpers.parent_domain(event.data).lower()
        return ".".join([s for s in query.split(".") if s != "_wildcard"])

    def parse_results(self, r, query=None):
        json = r.json()
        if json:
            for hostname in json:
                yield hostname

    def query(self, query, parse_fn=None, request_fn=None):
        if parse_fn is None:
            parse_fn = self.parse_results
        if request_fn is None:
            request_fn = self.request_url
        try:
            results = list(parse_fn(request_fn(query), query))
            if results:
                return results
            self.debug(f'No results for "{query}"')
        except Exception as e:
            self.info(f"Error retrieving results for {query}: {e}")
            self.trace()



from bbot.modules.base import BaseModule


class bucket_aws(BaseModule):
    watched_events = ["DNS_NAME", "STORAGE_BUCKET"]
    produced_events = ["STORAGE_BUCKET", "FINDING"]
    flags = ["active", "safe", "cloud-enum", "web-basic", "web-thorough"]
    meta = {"description": "Check for S3 buckets related to target"}
    options = {"max_threads": 10, "permutations": False}
    options_desc = {
        "max_threads": "Maximum number of threads for HTTP requests",
        "permutations": "Whether to try permutations",
    }
    scope_distance_modifier = 3

    cloud_helper_name = "aws"
    delimiters = ("", ".", "-")
    base_domains = ["s3.amazonaws.com"]
    regions = [None]
    supports_open_check = True

    def setup(self):
        self.buckets_tried = set()
        self.cloud_helper = getattr(self.helpers.cloud, self.cloud_helper_name)
        self.permutations = self.config.get("permutations", False)
        return True

    def filter_event(self, event):
        if event.type == "DNS_NAME" and event.scope_distance > 0:
            return False, "only accepts in-scope DNS_NAMEs"
        if event.type == "STORAGE_BUCKET":
            if f"cloud-{self.cloud_helper_name}" not in event.tags:
                return False, "bucket belongs to a different cloud provider"
        return True

    def handle_event(self, event):
        if event.type == "DNS_NAME":
            self.handle_dns_name(event)
        elif event.type == "STORAGE_BUCKET":
            self.handle_storage_bucket(event)

    def handle_dns_name(self, event):
        buckets = set()
        base = event.data
        stem = self.helpers.domain_stem(base)
        for b in [base, stem]:
            split = b.split(".")
            for d in self.delimiters:
                buckets.add(d.join(split))
        for bucket_name, url, tags in self.brute_buckets(buckets, permutations=self.permutations):
            self.emit_event({"name": bucket_name, "url": url}, "STORAGE_BUCKET", source=event, tags=tags)

    def handle_storage_bucket(self, event):
        url = event.data["url"]
        bucket_name = event.data["name"]
        if self.supports_open_check:
            description, tags = self._check_bucket_open(bucket_name, url)
            if description:
                event_data = {"host": event.host, "url": url, "description": description}
                self.emit_event(event_data, "FINDING", source=event, tags=tags)

        for bucket_name, url, tags in self.brute_buckets(
            [bucket_name], permutations=self.permutations, omit_base=True
        ):
            self.emit_event({"name": bucket_name, "url": url}, "STORAGE_BUCKET", source=event, tags=tags)

    def brute_buckets(self, buckets, permutations=False, omit_base=False):
        buckets = set(buckets)
        new_buckets = set(buckets)
        if permutations:
            for b in buckets:
                for mutation in self.helpers.word_cloud.mutations(b, cloud=False):
                    for d in self.delimiters:
                        new_buckets.add(d.join(mutation))
        if omit_base:
            new_buckets = new_buckets - buckets
        new_buckets = [b for b in new_buckets if self.valid_bucket_name(b)]
        futures = {}
        for base_domain in self.base_domains:
            for region in self.regions:
                for bucket_name in new_buckets:
                    url = self.build_url(bucket_name, base_domain, region)
                    future = self.submit_task(self._check_bucket_exists, bucket_name, url)
                    futures[future] = (bucket_name, url)
        for future in self.helpers.as_completed(futures):
            bucket_name, url = futures[future]
            existent_bucket, tags = future.result()
            if existent_bucket:
                yield bucket_name, url, tags

    def _check_bucket_exists(self, bucket_name, url):
        self.debug(f'Checking if bucket exists: "{bucket_name}"')
        return self.check_bucket_exists(bucket_name, url)

    def check_bucket_exists(self, bucket_name, url):
        response = self.helpers.request(url)
        tags = self.gen_tags_exists(response)
        status_code = getattr(response, "status_code", 404)
        existent_bucket = status_code != 404
        return (existent_bucket, tags)

    def _check_bucket_open(self, bucket_name, url):
        self.debug(f'Checking if bucket is misconfigured: "{bucket_name}"')
        return self.check_bucket_open(bucket_name, url)

    def check_bucket_open(self, bucket_name, url):
        response = self.helpers.request(url)
        tags = self.gen_tags_exists(response)
        status_code = getattr(response, "status_code", 404)
        content = getattr(response, "text", "")
        open_bucket = status_code == 200 and "Contents" in content
        msg = ""
        if open_bucket:
            msg = "Open storage bucket"
        return (msg, tags)

    def valid_bucket_name(self, bucket_name):
        valid = self.cloud_helper.is_valid_bucket(bucket_name)
        if valid and not self.helpers.is_ip(bucket_name):
            bucket_hash = hash(bucket_name)
            if not bucket_hash in self.buckets_tried:
                self.buckets_tried.add(bucket_hash)
                return True
        return False

    def build_url(self, bucket_name, base_domain, region):
        return f"https://{bucket_name}.{base_domain}"

    def gen_tags_exists(self, response):
        return set()

    def gen_tags_open(self, response):
        return set()



import queue
import logging
import threading
import traceback
from datetime import datetime
from queue import SimpleQueue, Full
from concurrent.futures import ThreadPoolExecutor

log = logging.getLogger("bbot.core.helpers.threadpool")

from .cache import CacheDict
from ...core.errors import ScanCancelledError


def pretty_fn(a):
    if callable(a):
        return a.__qualname__
    return a


class ThreadPoolSimpleQueue(SimpleQueue):
    def __init__(self, *args, **kwargs):
        self._executor = kwargs.pop("_executor", None)
        assert self._executor is not None, "Must specify _executor"

    def get(self, *args, **kwargs):
        work_item = super().get(*args, **kwargs)
        if work_item is not None:
            thread_id = threading.get_ident()
            self._executor._current_work_items[thread_id] = (work_item, datetime.now())
        return work_item


class PatchedThreadPoolExecutor(ThreadPoolExecutor):
    """
    This class exists only because of a bug in cpython where
    futures are not properly marked CANCELLED_AND_NOTIFIED:
        https://github.com/python/cpython/issues/87893
    """

    def shutdown(self, wait=True, *, cancel_futures=False):
        with self._shutdown_lock:
            self._shutdown = True
            if cancel_futures:
                # Drain all work items from the queue, and then cancel their
                # associated futures.
                while 1:
                    try:
                        work_item = self._work_queue.get_nowait()
                    except queue.Empty:
                        break
                    if work_item is not None:
                        if work_item.future.cancel():
                            work_item.future.set_running_or_notify_cancel()

            # Send a wake-up to prevent threads calling
            # _work_queue.get(block=True) from permanently blocking.
            self._work_queue.put(None)
        if wait:
            for t in self._threads:
                t.join()


class BBOTThreadPoolExecutor(PatchedThreadPoolExecutor):
    """
    Allows inspection of thread pool to determine which functions are currently executing
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._current_work_items = {}
        self._work_queue = ThreadPoolSimpleQueue(_executor=self)

    @property
    def threads_status(self):
        work_items = []
        for thread_id, (work_item, start_time) in sorted(self._current_work_items.items()):
            func = work_item.fn.__qualname__
            func_index = 0
            if work_item and not work_item.future.done():
                for i, f in enumerate(list(work_item.args)):
                    if callable(f):
                        func = f.__qualname__
                        func_index = i + 1
                    else:
                        break
                running_for = datetime.now() - start_time
                wi_args = list(work_item.args)[func_index:]
                wi_args = [pretty_fn(a) for a in wi_args]
                wi_args = str(wi_args).strip("[]")
                wi_kwargs = ", ".join(["{0}={1}".format(k, pretty_fn(v)) for k, v in work_item.kwargs.items()])
                func_with_args = f"{func}({wi_args}" + (f", {wi_kwargs}" if wi_kwargs else "") + ")"
                work_items.append(
                    (running_for, f"running for {int(running_for.total_seconds()):>3} seconds: {func_with_args}")
                )
        work_items.sort(key=lambda x: x[0])
        return [x[-1] for x in work_items]


class ThreadPoolWrapper:
    """
    Layers more granular control overtop of a shared thread pool
    Allows setting lower thread limits for modules, etc.
    """

    def __init__(self, executor, max_workers=None, qsize=None):
        self.executor = executor
        self.max_workers = max_workers
        self.max_qsize = qsize
        self.futures = set()

        self._num_tasks = 0
        self._task_count_lock = threading.Lock()

        self._lock = threading.RLock()
        self.not_full = threading.Condition(self._lock)

        try:
            self.executor._thread_pool_wrappers.append(self)
        except AttributeError:
            self.executor._thread_pool_wrappers = [self]

    def submit_task(self, callback, *args, **kwargs):
        """
        A wrapper around threadpool.submit()
        """
        block = kwargs.pop("_block", True)
        force = kwargs.pop("_force_submit", False)
        success = False
        with self.not_full:
            self.num_tasks_increment()
            try:
                if not force:
                    if not block:
                        if self.is_full or self.underlying_executor_is_full:
                            raise Full
                    else:
                        # wait until there's room
                        while self.is_full or self.underlying_executor_is_full:
                            self.not_full.wait()

                try:
                    # submit the job
                    future = self.executor.submit(self._execute_callback, callback, *args, **kwargs)
                    future.add_done_callback(self._on_future_done)
                    success = True
                    return future
                except RuntimeError as e:
                    raise ScanCancelledError(e)
            finally:
                if not success:
                    self.num_tasks_decrement()

    def _execute_callback(self, callback, *args, **kwargs):
        try:
            return callback(*args, **kwargs)
        finally:
            self.num_tasks_decrement()

    def _on_future_done(self, future):
        if future.cancelled():
            self.num_tasks_decrement()

    @property
    def num_tasks(self):
        with self._task_count_lock:
            return self._num_tasks

    def num_tasks_increment(self):
        with self._task_count_lock:
            self._num_tasks += 1

    def num_tasks_decrement(self):
        with self._task_count_lock:
            self._num_tasks = max(0, self._num_tasks - 1)
        for wrapper in self.executor._thread_pool_wrappers:
            try:
                with wrapper.not_full:
                    wrapper.not_full.notify()
            except RuntimeError:
                continue
            except Exception as e:
                log.warning(f"Unknown error in num_tasks_decrement(): {e}")
                log.trace(traceback.format_exc())

    @property
    def is_full(self):
        if self.max_workers is None:
            return False
        return self.num_tasks > self.max_workers

    @property
    def underlying_executor_is_full(self):
        return self.max_qsize is not None and self.qsize >= self.max_qsize

    @property
    def qsize(self):
        return self.executor._work_queue.qsize()

    def shutdown(self, *args, **kwargs):
        self.executor.shutdown(*args, **kwargs)

    @property
    def threads_status(self):
        return self.executor.threads_status


from concurrent.futures._base import FINISHED
from concurrent.futures._base import as_completed as as_completed_orig


def as_completed(*args, **kwargs):
    for f in as_completed_orig(*args, **kwargs):
        if f._state == FINISHED:
            yield f


class _Lock:
    def __init__(self, name):
        self.name = name
        self.lock = threading.Lock()

    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()


class NamedLock:
    """
    Returns a unique threading.Lock() based on a provided string

    Useful for preventing multiple operations from occuring on the same data in parallel
    E.g. simultaneous DNS lookups on the same hostname
    """

    def __init__(self, max_size=1000):
        self._cache = CacheDict(max_size=max_size)

    def get_lock(self, name):
        try:
            return self._cache.get(name)
        except KeyError:
            new_lock = _Lock(name)
            self._cache.put(name, new_lock)
            return new_lock


class TaskCounter:
    def __init__(self):
        self._value = 0
        self.lock = threading.Lock()

    def __enter__(self):
        with self.lock:
            self._value += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        with self.lock:
            self._value -= 1

    @property
    def value(self):
        with self.lock:
            return self._value



import re
import logging

log = logging.getLogger("bbot.helpers.cloud.provider")


class BaseCloudProvider:
    domains = []
    regexes = {}

    def __init__(self, parent_helper):
        self.parent_helper = parent_helper
        self.name = str(self.__class__.__name__).lower()
        self.dummy_module = self.parent_helper._make_dummy_module(f"{self.name}_cloud", _type="scan")
        self.bucket_name_regex = re.compile("^" + self.bucket_name_regex + "$", re.I)
        self.signatures = {}
        self.domain_regexes = []
        for domain in self.domains:
            self.domain_regexes.append(re.compile(r"^(?:[\w\-]+\.)*" + rf"{re.escape(domain)}$"))
        for event_type, regexes in self.regexes.items():
            self.signatures[event_type] = [re.compile(r, re.I) for r in regexes]

    @property
    def base_tags(self):
        return {f"cloud-{self.name}"}

    def excavate(self, event, http_body):
        base_kwargs = dict(source=event, tags=self.base_tags)

        # check for buckets in HTTP responses
        for event_type, sigs in self.signatures.items():
            found = set()
            for sig in sigs:
                for match in sig.findall(http_body):
                    kwargs = dict(base_kwargs)
                    kwargs["event_type"] = event_type
                    if not match in found:
                        found.add(match)
                        if event_type == "STORAGE_BUCKET":
                            self.emit_bucket(match, **kwargs)
                        else:
                            self.emit_event(**kwargs)

    def emit_bucket(self, match, **kwargs):
        bucket_name, bucket_domain = match
        kwargs["data"] = {"name": bucket_name, "url": f"https://{bucket_name}.{bucket_domain}"}
        self.emit_event(**kwargs)

    def emit_event(self, *args, **kwargs):
        excavate_module = self.parent_helper.scan.modules.get("excavate", None)
        if excavate_module:
            event = self.dummy_module.make_event(*args, **kwargs)
            excavate_module.emit_event(event)

    def is_valid_bucket(self, bucket_name):
        return self.bucket_name_regex.match(bucket_name)

    def tag_event(self, event):
        # tag the event if
        if event.host:
            # its host directly matches this cloud provider's domains
            if isinstance(event.host, str) and self.domain_match(event.host):
                event.tags.update(self.base_tags)
                return
            # or it has a CNAME that matches this cloud provider's domains
            for rh in event.resolved_hosts:
                if not self.parent_helper.is_ip(rh) and self.domain_match(rh):
                    event.tags.update(self.base_tags)

    def domain_match(self, s):
        for r in self.domain_regexes:
            if r.match(s):
                return True
        return False



from ..bbot_fixtures import *


def test_cli(monkeypatch, bbot_config):
    from bbot import cli

    monkeypatch.setattr(sys, "exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(os, "_exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(cli, "config", bbot_config)

    old_sys_argv = sys.argv

    home_dir = Path(bbot_config["home"])
    scans_home = home_dir / "scans"

    # basic scan
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "bbot",
            "-y",
            "-t",
            "127.0.0.1",
            "www.example.com",
            "-om",
            "human",
            "csv",
            "json",
            "-n",
            "test_cli_scan",
            "-c",
            "dns_resolution=False",
            "-o",
            "/tmp",
        ],
    )
    cli.main()

    scan_home = scans_home / "test_cli_scan"
    assert (scan_home / "wordcloud.tsv").is_file()
    assert (scan_home / "output.txt").is_file()
    assert (scan_home / "output.csv").is_file()
    assert (scan_home / "output.json").is_file()
    with open(scan_home / "output.csv") as f:
        lines = f.readlines()
        assert lines[0] == "Event type,Event data,IP Address,Source Module,Scope Distance,Event Tags\n"
        assert len(lines) > 1

    ip_success = False
    dns_success = False
    output_filename = scan_home / "output.txt"
    with open(output_filename) as f:
        lines = f.read().splitlines()
        for line in lines:
            if "[IP_ADDRESS]        \t127.0.0.1\tTARGET" in line:
                ip_success = True
            if "[DNS_NAME]          \twww.example.com\tTARGET" in line:
                dns_success = True
    assert ip_success and dns_success

    # show version
    monkeypatch.setattr("sys.argv", ["bbot", "--version"])
    cli.main()

    # show current config
    monkeypatch.setattr("sys.argv", ["bbot", "-y", "--current-config"])
    cli.main()

    # list modules
    monkeypatch.setattr("sys.argv", ["bbot", "-l"])
    cli.main()

    # list module options
    monkeypatch.setattr("sys.argv", ["bbot", "--help-all"])
    cli.main()

    # unpatch sys.argv
    monkeypatch.setattr("sys.argv", old_sys_argv)


def test_config_validation(monkeypatch, capsys, bbot_config):
    from bbot import cli
    from bbot.core.configurator import args

    monkeypatch.setattr(sys, "exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(os, "_exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(cli, "config", bbot_config)

    old_cli_config = args.cli_config

    # incorrect module option
    monkeypatch.setattr(args, "cli_config", ["bbot", "-c", "modules.ipnegibhor.num_bits=4"])
    cli.main()
    captured = capsys.readouterr()
    assert 'Could not find module option "modules.ipnegibhor.num_bits"' in captured.err
    assert 'Did you mean "modules.ipneighbor.num_bits"?' in captured.err

    # incorrect global option
    monkeypatch.setattr(args, "cli_config", ["bbot", "-c", "web_spier_distance=4"])
    cli.main()
    captured = capsys.readouterr()
    assert 'Could not find module option "web_spier_distance"' in captured.err
    assert 'Did you mean "web_spider_distance"?' in captured.err

    # unpatch cli_options
    monkeypatch.setattr(args, "cli_config", old_cli_config)


def test_module_validation(monkeypatch, capsys, bbot_config):
    from bbot.core.configurator import args

    monkeypatch.setattr(sys, "exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(os, "_exit", lambda *args, **kwargs: True)

    old_sys_argv = sys.argv

    # incorrect module
    monkeypatch.setattr(sys, "argv", ["bbot", "-m", "massdnss"])
    args.parser.parse_args()
    captured = capsys.readouterr()
    assert 'Could not find module "massdnss"' in captured.err
    assert 'Did you mean "massdns"?' in captured.err

    # incorrect excluded module
    monkeypatch.setattr(sys, "argv", ["bbot", "-em", "massdnss"])
    args.parser.parse_args()
    captured = capsys.readouterr()
    assert 'Could not find module "massdnss"' in captured.err
    assert 'Did you mean "massdns"?' in captured.err

    # incorrect output module
    monkeypatch.setattr(sys, "argv", ["bbot", "-om", "neoo4j"])
    args.parser.parse_args()
    captured = capsys.readouterr()
    assert 'Could not find output module "neoo4j"' in captured.err
    assert 'Did you mean "neo4j"?' in captured.err

    # incorrect flag
    monkeypatch.setattr(sys, "argv", ["bbot", "-f", "subdomainenum"])
    args.parser.parse_args()
    captured = capsys.readouterr()
    assert 'Could not find flag "subdomainenum"' in captured.err
    assert 'Did you mean "subdomain-enum"?' in captured.err

    # unpatch sys.argv
    monkeypatch.setattr("sys.argv", old_sys_argv)



import io
import os
import logging
import threading
import traceback
import subprocess
from contextlib import suppress

from .misc import smart_decode, rm_at_exit

log = logging.getLogger("bbot.core.helpers.command")


def run_live(self, command, *args, **kwargs):
    """
    Get live output, line by line, as a process executes
    You can also pass input=<iterator> and pipe data into the process' stdin
        - This lets you chain processes like so:

            ls_process = run_live(["ls", "/etc"])
            grep_process = run_live(["grep", "conf"], input=ls_process)
            for line in grep_process:
                log.success(line)

        - The above is roughly equivalent to:
            ls /etc | grep conf

    NOTE: STDERR is logged after the process exits, if its exit code is non-zero
        If you want to see it immediately, pass stderr=None
    """
    command, kwargs = self._prepare_command_kwargs(command, kwargs)
    _input = kwargs.pop("input", "")
    input_msg = ""
    if _input:
        kwargs["stdin"] = subprocess.PIPE
        input_msg = " (with stdin)"
    log.hugeverbose(f"run_live{input_msg}: {' '.join(command)}")
    try:
        with catch(subprocess.Popen, command, *args, **kwargs) as process:
            if _input:
                if type(_input) in (str, bytes):
                    _input = (_input,)
                self.feed_pipe(process.stdin, _input, text=False)
            for line in io.TextIOWrapper(process.stdout, encoding="utf-8", errors="ignore", line_buffering=True):
                yield line

            # surface stderr
            process.wait()
            if process.stderr and process.returncode != 0:
                stderr = smart_decode(process.stderr.read())
                if stderr:
                    command_str = " ".join(command)
                    log.warning(f"Stderr for {command_str}:\n\t{stderr}")
    except AttributeError as e:
        if not str(e) == "__enter__":
            raise


def run(self, command, *args, **kwargs):
    """
    Simple helper for running a command, and getting its output as a string
        process = run(["ls", "/tmp"])
        process.stdout --> "file1.txt\nfile2.txt"

    NOTE: STDERR is captured (not displayed) by default.
        If you want to see it, self.debug(process.stderr) or pass stderr=None
    """
    command, kwargs = self._prepare_command_kwargs(command, kwargs)
    if not "text" in kwargs:
        kwargs["text"] = True
    log.hugeverbose(f"run: {' '.join(command)}")
    result = catch(subprocess.run, command, *args, **kwargs)

    # surface stderr
    if result.stderr and result.returncode != 0:
        stderr = smart_decode(result.stderr)
        if stderr:
            command_str = " ".join(command)
            log.warning(f"Stderr for {command_str}:\n\t{stderr}")

    return result


def _prepare_command_kwargs(self, command, kwargs):
    if not "stdout" in kwargs:
        kwargs["stdout"] = subprocess.PIPE
    if not "stderr" in kwargs:
        kwargs["stderr"] = subprocess.PIPE
    sudo = kwargs.pop("sudo", False)

    command = [str(s) for s in command]
    env = kwargs.get("env", os.environ)
    if sudo and os.geteuid() != 0:
        self.depsinstaller.ensure_root()
        env["SUDO_ASKPASS"] = str((self.tools_dir / self.depsinstaller.askpass_filename).resolve())
        env["BBOT_SUDO_PASS"] = self.depsinstaller._sudo_password
        kwargs["env"] = env

        PATH = os.environ.get("PATH", "")
        LD_LIBRARY_PATH = os.environ.get("LD_LIBRARY_PATH", "")
        command = ["sudo", "-E", "-A", f"LD_LIBRARY_PATH={LD_LIBRARY_PATH}", f"PATH={PATH}"] + command
    return command, kwargs


def catch(callback, *args, **kwargs):
    try:
        return callback(*args, **kwargs)
    except FileNotFoundError as e:
        log.warning(f"{e} - missing executable?")
        log.trace(traceback.format_exc())
    except BrokenPipeError as e:
        log.warning(f"Error in subprocess: {e}")
        log.trace(traceback.format_exc())


def tempfile(self, content, pipe=True):
    """
    tempfile(["temp", "file", "content"]) --> Path("/home/user/.bbot/temp/pgxml13bov87oqrvjz7a")

    if "pipe" is True (the default), a named pipe is used instead of
    a true file, which allows python data to be piped directly into the
    process without taking up disk space
    """
    filename = self.temp_filename()
    rm_at_exit(filename)
    try:
        if type(content) not in (set, list, tuple):
            content = (content,)
        if pipe:
            os.mkfifo(filename)
            self.feed_pipe(filename, content, text=True)
        else:
            with open(filename, "w", errors="ignore") as f:
                for c in content:
                    line = f"{self.smart_decode(c)}\n"
                    f.write(line)
    except Exception as e:
        log.error(f"Error creating temp file: {e}")
        log.trace(traceback.format_exc())

    return filename


def _feed_pipe(self, pipe, content, text=True):
    try:
        if text:
            decode_fn = self.smart_decode
            newline = "\n"
        else:
            decode_fn = self.smart_encode
            newline = b"\n"
        try:
            if hasattr(pipe, "write"):
                try:
                    for c in content:
                        pipe.write(decode_fn(c) + newline)
                finally:
                    with suppress(Exception):
                        pipe.close()
            else:
                with open(pipe, "w") as p:
                    for c in content:
                        p.write(decode_fn(c) + newline)
        except BrokenPipeError:
            log.debug(f"Broken pipe in _feed_pipe()")
        except ValueError:
            log.debug(f"Error _feed_pipe(): {traceback.format_exc()}")
    except KeyboardInterrupt:
        self.scan.stop()
    except Exception as e:
        log.error(f"Error in _feed_pipe(): {e}")
        log.trace(traceback.format_exc())


def feed_pipe(self, pipe, content, text=True):
    t = threading.Thread(target=self._feed_pipe, args=(pipe, content), kwargs={"text": text}, daemon=True)
    t.start()


def tempfile_tail(self, callback):
    """
    Create a named pipe and execute a callback on each line
    """
    filename = self.temp_filename()
    rm_at_exit(filename)
    try:
        os.mkfifo(filename)
        t = threading.Thread(target=tail, args=(filename, callback), daemon=True)
        t.start()
    except Exception as e:
        log.error(f"Error setting up tail for file {filename}: {e}")
        log.trace(traceback.format_exc())
        return
    return filename


def tail(filename, callback):
    try:
        with open(filename, errors="ignore") as f:
            for line in f:
                line = line.rstrip("\r\n")
                callback(line)
    except Exception as e:
        log.error(f"Error tailing file {filename}: {e}")
        log.trace(traceback.format_exc())
